import torch
from torch_geometric.data import Data
from torch_geometric.utils import subgraph
import numpy as np
import os
import urllib.request

from .base import Dataset

class UniversalDataset(Dataset):
    """
    UniversalDataset class is designed to handle various types of graph data,
    enabling operations on datasets that include features, states, dynamic graphs, and edge attributes.

    Parameters
    ----------
    name : str, optional
           Name of the deataset to be loaded (Supported dataset only).
    root : str, optional
           Location of the dataset to be downloaded (Supported dataset only).
    x : torch.Tensor, optional
        Node features tensor of shape (num_samples, num_nodes, num_features). Represents the node features over multiple timesteps.
    states : torch.Tensor, optional
        Tensor representing various states of nodes, similar in structure to node features.
    y : torch.Tensor, optional
        Tensor representing target labels or values for each node, structured similar to node features.
    graph : torch.Tensor or scipy.sparse matrix, optional
        Static graph structure as an adjacency matrix.
    dynamic_graph : torch.Tensor, optional
        Dynamic graph information over time, providing evolving adjacency matrices.
    edge_index : torch.LongTensor, optional
        Tensor containing edge indices, typically of shape (2, num_edges), for defining which nodes are connected.
    edge_weight : torch.Tensor, optional
        Edge weights corresponding to the edge_index, providing the strength or capacity of connections.
    edge_attr : torch.Tensor, optional
        Attributes or features for each edge, aligned with the structure defined in edge_index.
    """
    def __init__(
                self,
                name=None,
                root='./',
                x=None, # timestep * Num nodes * features 
                states=None,
                y=None,
                graph=None,
                dynamic_graph=None,
                edge_index=None,
                edge_weight=None,
                edge_attr = None
                ):
        super().__init__()

        self.x = x # N*D; L*D; L*N*D; 
        self.y = y # N*1; L*1; L*N*1
        self.graph = graph
        self.dynamic_graph = dynamic_graph
        self.edge_index = edge_index # None; 2*Links; L*2*Links
        self.edge_weight = edge_weight # same as edge_index
        self.edge_attr = edge_attr # same as edge_index
        self.states = states # same as x
        self.output_dim = None

        if name is not None:
            if not os.path.exists(root):
                    os.mkdir(root)
                    self.root = root
            if name == 'JHU_covid':
                if not os.path.exists(f"{root}/JHU_covid.pt"):
                    print("downloading JHU Covid Dataset")
                    url_feature = "https://drive.google.com/uc?export=download&id=13i9OpTweVYOvSOET-91-ZNVMJRURwxb0"
                    urllib.request.urlretrieve(url_feature, f"{root}/JHU_covid.pt")
                data = torch.load(f"{root}/JHU_covid.pt")
                for k, v in data.items():
                    if k=='features':
                        self.x = v.float()
                    if k=='feature_names':
                        self.features=v
                    if k=='index':
                        self.index=v

            elif name == 'Measles':
                if not os.path.exists(f"{root}/measles.pt"):
                    print("downloading Measles Dataset")
                    url_feature = "https://drive.google.com/uc?export=download&id=1-kLtvyUGN_mYJL5MgafEAd30KyfSmE4U"
                    urllib.request.urlretrieve(url_feature, f"{root}/measles.pt")

                data = torch.load(f"{root}/measles.pt")
                self.x = torch.FloatTensor(np.array(data['weekly_infection'])).T
                self.features = list(data['weekly_infection'].columns)
                self.anual_population = data['anual_population']
                self.anual_birth = data['anual_birth']
                self.coordinates = data['coordinates']

            elif name == 'Tycho_v1':
                if not os.path.exists(f"{root}/Tycho_v1.pt"):
                    print("downloading Tycho_v1 Dataset")
                    url_feature = "https://drive.google.com/uc?export=download&id=13gHDO6Rh5gZwqo8MLDyyiLtPCd9gD0nD"
                    urllib.request.urlretrieve(url_feature, f"{root}/Tycho_v1.pt")
                data = torch.load(f"{root}/Tycho_v1.pt")
                self.features = list(data.keys())
                self.x = []
                for v in data.values():
                    self.x.append(v.float())  

            elif name.split('_')[0] == 'Covid':
                country = name.split('_')[1]
                if not os.path.exists(f"{root}/covid_static.pt"):
                    print("downloading Covid Static Dataset") 
                    url_feature = "https://drive.google.com/uc?export=download&id=1-l1yVWlKxB0VLwj0IqJRAbruUvyrBUUI"
                    urllib.request.urlretrieve(url_feature, f"{root}/covid_static.pt")
                if not os.path.exists(f"{root}/covid_dynamic.pt"):
                    print("downloading Covid Dynamic Dataset") 
                    url_feature = "https://drive.google.com/uc?export=download&id=1-lUm1uaSDgbto2IV8aalZ3_lXh7LS7z3"
                    urllib.request.urlretrieve(url_feature, f"{root}/covid_dynamic.pt")

                try:
                    static_data = torch.load(f"{root}/covid_static.pt")
                    data = static_data[country]
                    self.graph = data['graph']
                except:
                    dynamic_data = torch.load(f"{root}/covid_dynamic.pt")
                    data = dynamic_data[country]
                    self.dynamic_graph = data['Dynamic_graph']

                self.x = data['features']
                self.features = data['feature_names']
                self.timestamp = data['time_stamp'] 
            else:
                raise ValueError("Dataset Not Found!")

        self.transforms = None

        if self.y is not None:
            self.output_dim = self.y.shape
        
        if self.graph is not None and self.edge_index is None:
            sparse_adj = torch.Tensor(self.graph).float().to_sparse()
            self.edge_index = sparse_adj.indices()
            self.edge_weight = sparse_adj.values()
    
    def get_transformed(self, input_data=None, transformations=None):
        if transformations is None:
            if self.transforms is not None:
                transformations = self.transforms
        if input_data is None:
            input_data = {"features": self.x, 
                        "target": self.y,
                        "graph": self.graph, 
                        "dynamic_graph": self.dynamic_graph, 
                        "states": self.states
                        }
        if transformations is not None:
            input_data = transformations(input_data)

        return input_data # input_data['features'], input_data['target'], input_data['graph'], input_data['dynamic_graph'], input_data['states']

    def get_split(self, inputs, idx1, idx2):
        if inputs is None:
            return None, None, None
        
        train_outputs = inputs[:idx1, ...]
        val_outputs = inputs[idx1:idx2, ...]
        test_outputs = inputs[idx2:, ...]

        return train_outputs, val_outputs, test_outputs
    
    def ganerate_splits(self, train_rate=0.6, val_rate=0.1):
        dataset = {
            "features": self.x,
            "target": self.y,
            "graph": self.graph,
            "dynamic_graph": self.dynamic_graph,
            "states": self.states
        }

        transformed_dataset = self.get_transformed(dataset)

        adj_static = self.graph

        split_line1 = int(self.x.shape[0] * train_rate)
        split_line2 = int(self.x.shape[0] * (train_rate + val_rate))

        train_features, val_features, test_features = self.get_split(transformed_dataset['features'], split_line1, split_line2)
        train_target, val_target, test_target = self.get_split(transformed_dataset['target'], split_line1, split_line2)
        train_states, val_states, test_states = self.get_split(transformed_dataset['states'], split_line1, split_line2)
        train_dynamic_adj, val_dynamic_adj, test_dynamic_adj = self.get_split(transformed_dataset['dynamic_graph'], split_line1, split_line2)
        
        train_dataset = {
            "features": train_features,
            "target": train_target,
            "graph": adj_static,
            "dynamic_graph": train_dynamic_adj,
            "states": train_states
        }
        val_dataset = {
            "features": val_features,
            "target": val_target,
            "graph": adj_static,
            "dynamic_graph": val_dynamic_adj,
            "states": val_states
        }
        test_dataset = {
            "features": test_features,
            "target": test_target,
            "graph": adj_static,
            "dynamic_graph": test_dynamic_adj,
            "states": test_states
        }

        return train_dataset, val_dataset, test_dataset

        
    def __getitem__(self, index):
        batch_y = None if self.y is None else self.y[index].unsqueeze(0)
        batch_dynamic = None if self.dynamic_graph is None else self.dynamic_graph[index].unsqueeze(0)

        if self.y is not None:
            return Data(x=self.x[index], y=batch_y, edge_index=self.edge_index, edge_attr=self.edge_weight, dynamic_graph=batch_dynamic)
        else:
            return Data(x=self.x[index], edge_index=self.edge_index, edge_attr=self.edge_weight, dynamic_graph=batch_dynamic)
        
    def __len__(self):
        return len(self.x)

    def download(self):
        pass

    def save(self):
        pass

    def generate_dataset(self, X = None, Y = None, states = None, dynamic_adj = None, lookback_window_size = 1, horizon_size = 1, permute = False):
        """
        Takes node features for the graph and divides them into multiple samples
        along the time-axis by sliding a window of size (num_timesteps_input+
        num_timesteps_output) across it in steps of 1.
        :param X: Node features of shape (num_vertices, num_features,
        num_timesteps)
        :return:
            - Node features divided into multiple samples. Shape is
            (num_samples, num_vertices, num_features, num_timesteps_input).
            - Node targets for the samples. Shape is
            (num_samples, num_vertices, num_features, num_timesteps_output).
        """
        if X is None:
            X = self.x
        if Y is None:
            Y = self.y

        indices = [(i, i + (lookback_window_size + horizon_size)) for i in range(X.shape[0] - (lookback_window_size + horizon_size) + 1)]
        target = []
        for i, j in indices:
            target.append(Y[i + lookback_window_size: j])
        
        targets = torch.stack(target) if len(target) > 0 else torch.Tensor([[[]]])
        if permute:
            targets = targets.transpose(1,2)

        input_list = [[None, X], [None, states], [None, dynamic_adj]]
        for m, inputs in enumerate(input_list):
            if inputs[1] is not None:
                tmp = []
                for i, j in indices:
                    tmp.append(inputs[1][i: i + lookback_window_size])
                input_list[m][0] = torch.stack(tmp) if len(tmp) else torch.Tensor([[[]]])
                if permute:
                    inputs[0] = inputs[0].transpose(1,2)

        return input_list[0][0], targets, input_list[1][0], input_list[2][0]

    
    def load_toy_dataset(self):
        if not os.path.exists("./datasets"):
            os.mkdir("./datasets/")
        if not os.path.exists("./datasets/features.npy"):
            print("downloading toy features")
            url_feature = "https://drive.google.com/uc?export=download&id=10VRjabU1m0pluQKOTQ-GKxF3sK9bLFYF"
            urllib.request.urlretrieve(url_feature, './datasets/features.npy')

        if not os.path.exists("./datasets/graphs.npy"):
            print("downloading toy graphs")
            url_graph = "https://drive.google.com/uc?export=download&id=10ZR4k19wdWXQdPN53Tz3QAQxBiZUXEPr"
            
            urllib.request.urlretrieve(url_graph, './datasets/graphs.npy')

        data1 = np.load(f"./datasets/graphs.npy")
        data2 = np.load(f"./datasets/features.npy",  allow_pickle= True)

        self.x = torch.FloatTensor(data2.tolist()['node'])
        self.y = torch.FloatTensor(data2.tolist()['node'])[:,:,0]
       
        self.dynamic_graph = torch.FloatTensor(data2.tolist()['od'])
        self.states = torch.FloatTensor(data2.tolist()['SIR'])

        self.graph = torch.FloatTensor(data1)
        self.edge_index = torch.FloatTensor(data1).to_sparse_coo().indices()
        self.edge_weight = torch.FloatTensor(data1).to_sparse_coo().values()
    


    

# class SpatialDataset(Dataset):
#     def __init__(   self,
#                     x = None,
#                     states = None,
#                     y = None,
#                     adj_m=None,
#                     edge_index = None,
#                     edge_attr = None
#                 ):
        
#         super().__init__()

#         self.x = x # N*D; L*D; L*N*D; 
#         self.y = y # N*1; L*1; L*N*1
#         self.adj_m = adj_m
#         self.edge_index = edge_index # None; 2*Links; L*2*Links
#         self.edge_attr = edge_attr # same as edge_index
#         self.states = states # same as x
#         self.output_dim = None
        
        
#         assert self.x is not None, "Input should not be NoneType!"
#         if self.y is not None:
#             assert len(self.x)==len(self.y), "Input and Output dim do not match!"
#             self.output_dim = self.y.shape
            
        
#         if self.edge_index is None and self.adj_m is None:
#             raise ValueError("There is no graph in Dataset, or you may specify your graph with parameter adj_m or edge_index!")
            
#         if self.adj_m is not None and self.edge_index is not None:
#             raise ValueError("There may be conflicts between your parameter adj_m and edge_index!")
        
#         if self.adj_m is not None and self.edge_index is None:
#             rows, cols = torch.where(self.adj_m != 0)
#             weights = self.adj_m[rows, cols]

#             edge_index = torch.stack([rows.long(), cols.long()], dim=0)
#             edge_weight = weights.clone().detach()
            
#             self.edge_index, self.edge_attr = edge_index, edge_weight
            
        
#         if self.adj_m is None and self.edge_index is not None:
#             num_nodes = self.x.shape[1] # or specify?
#             if self.edge_attr is None:
#                 self.edge_attr = torch.ones(self.edge_index.shape[1], 1)
#             if len(self.edge_attr.shape) == 1:
#                 self.edge_attr = torch.reshape(self.edge_attr, (self.edge_attr.shape[0], 1))
#             adj_matrix = torch.zeros((num_nodes, num_nodes, self.edge_attr.shape[1]))
#             for i in range(self.edge_index.shape[1]):
#                 src = self.edge_index[0, i]
#                 dest = self.edge_index[1, i]
#                 adj_matrix[src, dest] = self.edge_attr[i]
#             self.adj_m = adj_matrix.squeeze()
#             #print(self.adj_m.shape)
            
        
#     def __getitem__(self, index):
#         if self.y is not None:
#             return Data(x=self.x[index], y=self.y[index], edge_index=self.edge_index, edge_attr=self.edge_attr, adj_m=self.adj_m)
#         else:
#             return Data(x=self.x[index], edge_index=self.edge_index, edge_attr=self.edge_attr, adj_m=self.adj_m)
        
    
    
#     def __len__(self):
#         return len(self.x)
        






from .SIR import *
from . import (
    ARIMA,
    Dlinear,
    EINN,
    GRU,
    LSTM,
    # XGB,
    CNN
)

from .ARIMA import VARMAXModel, ARIMAModel
from .Dlinear import DlinearModel
# from .EINN import EINN
from .GRU import GRUModel
from .LSTM import LSTMModel
# from .XGB import XGBModel
from .CNN import CNNModel


# # this code is adopted from 

import torch
import torch.nn as nn
from torch_geometric.nn.inits import glorot, zeros
from typing import Union, Tuple
from torch_geometric.typing import OptPairTensor, Adj, Size

from torch import Tensor, randn
from torch.nn import Linear, Parameter
import torch.nn.functional as F
from torch_sparse import SparseTensor, matmul
from torch_geometric.nn.conv import MessagePassing
from torch import matmul

from .base import BaseModel


class WeightedSAGEConv(MessagePassing):
    r"""The GraphSAGE operator from the `"Inductive Representation Learning on
    Large Graphs" <https://arxiv.org/abs/1706.02216>`_ paper

    .. math::
        \mathbf{x}^{\prime}_i = \mathbf{W}_1 \mathbf{x}_i + \mathbf{W_2} \cdot
        \mathrm{mean}_{j \in \mathcal{N(i)}} \mathbf{x}_j

    Args:
        in_channels (int or tuple): Size of each input sample. A tuple
            corresponds to the sizes of source and target dimensionalities.
        out_channels (int): Size of each output sample.
        normalize (bool, optional): If set to :obj:`True`, output features
            will be :math:`\ell_2`-normalized, *i.e.*,
            :math:`\frac{\mathbf{x}^{\prime}_i}
            {\| \mathbf{x}^{\prime}_i \|_2}`.
            (default: :obj:`False`)
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.MessagePassing`.
    """
    def __init__(self, in_channels: Union[int, Tuple[int, int]],
                 out_channels: int, normalize: bool = False,
                 bias: bool = True, weighted: bool = True, edge_count: int = 423, **kwargs):  # yapf: disable
        super(WeightedSAGEConv, self).__init__(aggr='mean', **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.normalize = normalize
        self.weighted = weighted

        if isinstance(in_channels, int):
            in_channels = (in_channels, in_channels)

        self.lin_l = Linear(in_channels[0], out_channels, bias=bias)
        self.lin_r = Linear(in_channels[1], out_channels, bias=False)

        self.lin_m = Linear(in_channels[0], in_channels[0], bias=bias)
        self.lin_ew = Linear(1, 1, bias=bias)

        self.lin_edge_feature_weighting = Linear(3, 1, bias=bias)

        self.edge_attr = Parameter(randn(edge_count, 1))

        self.reset_parameters()

    def reset_parameters(self):
        self.lin_l.reset_parameters()
        self.lin_r.reset_parameters()

    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj, edge_weight: Tensor = None,
                size: Size = None) -> Tensor:

        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor)
        out = self.propagate(edge_index=edge_index, size=size)
        out = self.lin_l(out)

        x_r = x[1]
        if x_r is not None:
            out += self.lin_r(x_r)

        if self.normalize:
            out = F.normalize(out, p=2., dim=-1)

        return out

    def message(self, x_i: Tensor, x_j: Tensor, edge_weight) -> Tensor:
        # reduced_weight = self.lin_edge_feature_weighting(edge_weight)
        return x_i * edge_weight # self.lin_edge_feature_weighting(edge_weight)
        if not self.weighted:
            return x_j
        out = self.lin_m(x_j - x_i)
        weight = self.lin_ew(self.edge_attr)
        return out if self.edge_attr is None else out * weight

    def message_and_aggregate(self, adj_t: SparseTensor,
                              x: OptPairTensor) -> Tensor:
        adj_t = adj_t.set_value(None, layout=None)
        return matmul(adj_t, x[0], reduce=self.aggr)

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.in_channels,
                                   self.out_channels)

class LSTM(torch.nn.Module):
    # This is an adaptation of torch_geometric_temporal.nn.GConvLSTM, with ChebConv replaced by the given model.
    """
    Args:
        in_channels (int): Number of input features.
        out_channels (int): Number of output features.
        normalization (str, optional): The normalization scheme for the graph
            Laplacian (default: :obj:`"sym"`):

            1. :obj:`None`: No normalization
            :math:`\mathbf{L} = \mathbf{D} - \mathbf{A}`

            2. :obj:`"sym"`: Symmetric normalization
            :math:`\mathbf{L} = \mathbf{I} - \mathbf{D}^{-1/2} \mathbf{A}
            \mathbf{D}^{-1/2}`

            3. :obj:`"rw"`: Random-walk normalization
            :math:`\mathbf{L} = \mathbf{I} - \mathbf{D}^{-1} \mathbf{A}`

            You need to pass :obj:`lambda_max` to the :meth:`forward` method of
            this operator in case the normalization is non-symmetric.
            :obj:`\lambda_max` should be a :class:`torch.Tensor` of size
            :obj:`[num_graphs]` in a mini-batch scenario and a
            scalar/zero-dimensional tensor when operating on single graphs.
            You can pre-compute :obj:`lambda_max` via the
            :class:`torch_geometric.transforms.LaplacianLambdaMax` transform.
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
        module (torch.nn.Module, optional): The layer or set of layers used to calculate each gate.
            Could also be a lambda function returning a torch.nn.Module when given the parameters in_channels: int, out_channels: int, and bias: bool
    """
    def __init__(self, in_channels: int, out_channels: int, bias: bool=True, module=WeightedSAGEConv):
        super(LSTM, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.bias = bias
        self.module = module
        self._create_parameters_and_layers()
        self._set_parameters()


    def _create_input_gate_parameters_and_layers(self):

        self.conv_x_i = self.module(in_features=self.in_channels,
                                 out_features=self.out_channels,
                                 bias=self.bias)

        self.conv_h_i = self.module(in_features=self.out_channels,
                                 out_features=self.out_channels,
                                 bias=self.bias)

        self.w_c_i = Parameter(torch.Tensor(1, self.out_channels))
        self.b_i = Parameter(torch.Tensor(1, self.out_channels))


    def _create_forget_gate_parameters_and_layers(self):

        self.conv_x_f = self.module(in_features=self.in_channels,
                                 out_features=self.out_channels,
                                 bias=self.bias)

        self.conv_h_f = self.module(in_features=self.out_channels,
                                 out_features=self.out_channels,
                                 bias=self.bias)

        self.w_c_f = Parameter(torch.Tensor(1, self.out_channels))
        self.b_f = Parameter(torch.Tensor(1, self.out_channels))


    def _create_cell_state_parameters_and_layers(self):

        self.conv_x_c = self.module(in_features=self.in_channels,
                                 out_features=self.out_channels,
                                 bias=self.bias)

        self.conv_h_c = self.module(in_features=self.out_channels,
                                 out_features=self.out_channels,
                                 bias=self.bias)

        self.b_c = Parameter(torch.Tensor(1, self.out_channels))


    def _create_output_gate_parameters_and_layers(self):

        self.conv_x_o = self.module(in_features=self.in_channels,
                                 out_features=self.out_channels,
                                 bias=self.bias)

        self.conv_h_o = self.module(in_features=self.out_channels,
                                 out_features=self.out_channels,
                                 bias=self.bias)

        self.w_c_o = Parameter(torch.Tensor(1, self.out_channels))
        self.b_o = Parameter(torch.Tensor(1, self.out_channels))


    def _create_parameters_and_layers(self):
        self._create_input_gate_parameters_and_layers()
        self._create_forget_gate_parameters_and_layers()
        self._create_cell_state_parameters_and_layers()
        self._create_output_gate_parameters_and_layers()


    def _set_parameters(self):
        glorot(self.w_c_i)
        glorot(self.w_c_f)
        glorot(self.w_c_o)
        zeros(self.b_i)
        zeros(self.b_f)
        zeros(self.b_c)
        zeros(self.b_o)


    def _set_hidden_state(self, X, H):
        if H is None:
            H = torch.zeros(X.shape[0], self.out_channels)
        return H


    def _set_cell_state(self, X, C):
        if C is None:
            C = torch.zeros(X.shape[0], self.out_channels)
        return C


    def _calculate_input_gate(self, X, edge_index, edge_weight, H, C):
        I = self.conv_x_i(X, edge_index, edge_weight)
        I = I + self.conv_h_i(H, edge_index, edge_weight)
        I = I + (self.w_c_i*C)
        I = I + self.b_i
        I = torch.sigmoid(I)
        return I


    def _calculate_forget_gate(self, X, edge_index, edge_weight, H, C):
        F = self.conv_x_f(X, edge_index, edge_weight)
        F = F + self.conv_h_f(H, edge_index, edge_weight)
        F = F + (self.w_c_f*C)
        F = F + self.b_f
        F = torch.sigmoid(F)
        return F


    def _calculate_cell_state(self, X, edge_index, edge_weight, H, C, I, F):
        T = self.conv_x_c(X, edge_index, edge_weight)
        T = T + self.conv_h_c(H, edge_index, edge_weight)
        T = T + self.b_c
        T = torch.tanh(T)
        C = F*C + I*T
        return C

    def _calculate_output_gate(self, X, edge_index, edge_weight, H, C):
        O = self.conv_x_o(X, edge_index, edge_weight)
        O = O + self.conv_h_o(H, edge_index, edge_weight)
        O = O + (self.w_c_o*C)
        O = O + self.b_o
        O = torch.sigmoid(O)
        return O


    def _calculate_hidden_state(self, O, C):
        H = O * torch.tanh(C)
        return H


    def forward(self, X: torch.FloatTensor, edge_index: torch.LongTensor, edge_weight: torch.FloatTensor=None,
                H: torch.FloatTensor=None, C: torch.FloatTensor=None) -> torch.FloatTensor:
        """
        Making a forward pass. If edge weights are not present the forward pass
        defaults to an unweighted graph. If the hidden state and cell state
        matrices are not present when the forward pass is called these are
        initialized with zeros.

        Arg types:
            * **X** *(PyTorch Float Tensor)* - Node features.
            * **edge_index** *(PyTorch Long Tensor)* - Graph edge indices.
            * **edge_weight** *(PyTorch Long Tensor, optional)* - Edge weight vector.
            * **H** *(PyTorch Float Tensor, optional)* - Hidden state matrix for all nodes.
            * **C** *(PyTorch Float Tensor, optional)* - Cell state matrix for all nodes.

        Return types:
            * **H** *(PyTorch Float Tensor)* - Hidden state matrix for all nodes.
            * **C** *(PyTorch Float Tensor)* - Cell state matrix for all nodes.
        """
        H = self._set_hidden_state(X, H)
        C = self._set_cell_state(X, C)
        I = self._calculate_input_gate(X, edge_index, edge_weight, H, C)
        F = self._calculate_forget_gate(X, edge_index, edge_weight, H, C)
        C = self._calculate_cell_state(X, edge_index, edge_weight, H, C, I, F)
        O = self._calculate_output_gate(X, edge_index, edge_weight, H, C)
        H = self._calculate_hidden_state(O, C)
        return H, C


class GraphLinear(torch.nn.Linear):
    """This is the exact same as torch.nn.Linear,
    except that it can take edge_index, edge_attr and do nothing with them.
    Makes it interchangeable with graph neural network modules."""

    def forward(self, input, edge_index, edge_attr):
        return super(GraphLinear, self).forward(input)


class GraphLSTM(BaseModel):
    """
    Base class for Recurrent Neural Networks (LSTM or GRU).
    Initialization to this class contains all variables for variation of the model.
    Consists of one of the above RNN architectures followed by an optional GNN on the final hidden state.
    Parameters:
        node_features: int - number of features per node
        output: int - length of the output vector on each node
        dim: int - number of features of embedding for each node
        module: torch.nn.Module - to be used in the LSTM to calculate each gate
    """
    def __init__(self, 
                num_nodes, 
                num_features,
                num_timesteps_input,
                num_timesteps_output, 
                dim=32, 
                module=GraphLinear, 
                rnn=LSTM, 
                gnn=WeightedSAGEConv, 
                gnn_2=WeightedSAGEConv, 
                rnn_depth=1, 
                name="RNN", 
                skip_connection=True):
        super(GraphLSTM, self).__init__()
        self.n_feat = num_features
        self.n_in = num_timesteps_input
        self.m = num_nodes
        self.dim = dim
        self.rnn_depth = rnn_depth
        self.name = name
        self.skip_connection = skip_connection

        # Ensure that matrix multiplication sizes match up based on whether GNNs and RNN are used
        if gnn:
            if skip_connection:
                self.gnn = gnn(num_features, dim)
            else:
                self.gnn = gnn(num_features, dim * 2)
            if rnn:
                if skip_connection:
                    self.recurrent = rnn(dim, dim, module=module)
                else:
                    self.recurrent = rnn(dim * 2, dim * 2, module=module)
            else:
                self.recurrent = None
        else:
            self.gnn = None
            if rnn:
                self.recurrent = rnn(num_features, dim, module=module)
            else:
                self.recurrent = None
        if gnn_2:
            if gnn:
                self.gnn_2 = gnn_2(dim * 2, dim * 2)
            else:
                self.gnn_2 = gnn_2(dim + num_features, dim * 2)
        else:
            self.gnn_2 = None

        self.lin1 = torch.nn.Linear(2 * dim, dim)
        self.lin2 = torch.nn.Linear(dim, num_timesteps_output)
        self.act1 = torch.nn.ReLU()
        self.act2 = torch.nn.ReLU()

    def forward(self, X, adj, states=None, dynamic_adj=None, h=None, c=None):
        # Get data from snapshot
        x, edge_index = X, adj.to_sparse()

        # First GNN Layer
        if self.gnn:
            x = self.gnn(x, edge_index)
            x = F.relu(x)

        # Initialize hidden and cell states if None
        current_dim = self.dim
        if not self.skip_connection:
            current_dim = self.dim * 2
        if h is None:
            h = torch.zeros(x.shape[0], current_dim)
        if c is None:
            c = torch.zeros(x.shape[0], current_dim)

        # RNN Layer
        if self.recurrent:
            for i in range(self.rnn_depth):
                h, c = self.recurrent(x, edge_index, edge_attr, h, c)

        # Skip connection from first GNN
        if self.skip_connection:
            x = torch.cat((x, h), 1)
        else:
            x = h

        # Second GNN Layer
        if self.gnn_2:
            x = self.gnn_2(x, edge_index, edge_attr)

        # Readout and activation layers
        x = self.lin1(x)
        # x = self.act1(x)
        x = self.lin2(x)
        # x = self.act2(x)

        return x, h, c
    

    def initialize(self):
        for layer in self.children():
            if hasattr(layer, 'reset_parameters'):
                layer.reset_parameters()


# from . import (
#             STGCN, 
#             ATMGNN, 
#             CNNRNN_Res, 
#             STAN, 
#             ColaGNN, 
#             EpiColaGNN, 
#             GraphLSTM, 
#             MepoGNN, 
#             EpiGNN, 
#             DASTGN, 
#             HierST, 
#             DCRNN,
#             GraphWaveNet,
#             DMP,
#             NetworkSIR,
# )

from .STGCN import STGCN
from .ATMGNN import ATMGNN
from .CNNRNN_Res import CNNRNN_Res
from .STAN import STAN
from .ColaGNN import ColaGNN
from .EpiColaGNN import EpiColaGNN
# from .GraphLSTM import GraphLSTM
from .MepoGNN import MepoGNN
from .EpiGNN import EpiGNN
from .DASTGN import DASTGN
# from .HierST import HierST
from .DCRNN import DCRNN
from .GraphWaveNet import GraphWaveNet
from .DMP import DMP
from .NetworkSIR import NetSIR




import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init

from .base import BaseModel

class CNNModel(BaseModel):
    """
        Single-layer Gated Recurrent Unit (GRU) Network

        Parameters
        ----------
        num_features : int
            Number of features in the input data.
        num_timesteps_input : int
            Number of input timesteps.
        num_timesteps_output : int
            Number of output timesteps to predict.
        nhid : int, optional
            Number of hidden units in the GRU layer. Default: 256.
        dropout : float, optional
            Dropout rate for the GRU layer. Default: 0.5.
        use_norm : bool, optional
            Whether to use Layer Normalization after the GRU layer. Default: False.

        Returns
        -------
        torch.Tensor
            A tensor of shape (batch_size, num_timesteps_output) representing the predicted values for the future timesteps.
            Each element corresponds to a predicted value for a future timestep.
            
        """
    def __init__(self, 
                 num_features, 
                 num_timesteps_input, 
                 num_timesteps_output, 
                 dropout=0.5, 
                 conv1_config={'hid': 16, 'kernel': 3, 'stride': 1, 'padding': 1}, 
                 conv2_config={'hid': 32, 'kernel': 3, 'stride': 1, 'padding': 1},
                 maxpool_config={'kernel': 2, 'stride': 2, 'padding': 0}, 
                 linear_hid=128, 
                 device='cpu'):
        super(CNNModel, self).__init__(device=device)
        self.num_features = num_features
        self.num_timesteps_input = num_timesteps_input
        self.num_timesteps_output = num_timesteps_output
        self.dropout = dropout

        self.conv1 = nn.Conv1d(in_channels=self.num_features,
                               out_channels=conv1_config['hid'],
                               kernel_size=conv1_config['kernel'],
                               stride=conv1_config['stride'],
                               padding=conv1_config['padding']) # first convolutional layer
        self.conv2 = nn.Conv1d(in_channels=conv1_config['hid'],
                               out_channels=conv2_config['hid'],
                               kernel_size=conv2_config['kernel'],
                               stride=conv2_config['stride'],
                               padding=conv2_config['padding']) # second convolutional layer
        self.pool = nn.MaxPool1d(kernel_size=maxpool_config['kernel'],
                                 stride=maxpool_config['stride'],
                                 padding=maxpool_config['padding']) # Max Pooling
        self.fc1 = nn.Linear(conv2_config['hid'] * (num_timesteps_input // 4), linear_hid) # first linear layer
        self.fc2 = nn.Linear(linear_hid, num_timesteps_output) # second linear layer

        self.dropout = nn.Dropout(0.5)


    def forward(self, x):
        """
        Parameters
        ----------
        x : torch.Tensor
            The input tensor for the model. Expected shape is 
            (batch_size, num_timesteps_input, num_features), where
            `batch_size` is the number of samples in the batch,
            `num_timesteps_input` is the number of input timesteps,
            and `num_features` is the number of features for each timestep.

        Returns
        -------
        torch.Tensor
            The output of the model, a tensor of shape 
            (batch_size, num_timesteps_output), representing the predicted values 
            for the future timesteps. Each element corresponds to a predicted value 
            for a future timestep.
        """
        # feature extrction using convolution layers
        x = x.permute([0,2,1])
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))

        # flatten and perform prediction using fully connected layers
        x = x.view(x.size(0), -1)
        x = self.dropout(x)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        
        return x

    def initialize(self):
        for m in self.modules():
            if isinstance(m, nn.Conv1d) or isinstance(m, nn.Linear):
                init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    init.constant_(m.bias, 0)
        
        


# import xgboost as xgb
# from sklearn.metrics import mean_squared_error
# import torch.nn.init as init
# import torch

# class XGBModel:
#     """
#         Extreme Gradient Boosting (XGBoost) Model
        
#         Parameters
#         ----------
#         num_features : int
#             Number of features in each timestep of the input data.
#         num_timesteps_input : int
#             Number of timesteps considered for each input sample.
#         num_timesteps_output : int
#             Number of output timesteps to predict.
#         n_estimators : int, optional
#             Number of gradient boosted trees. Equivalent to the number of boosting rounds. Default: 40.
#         learning_rate : float, optional
#             Step size shrinkage used in update to prevents overfitting. Default: 0.1.
#         max_depth : int, optional
#             Maximum depth of a tree. Increasing this value will make the model more complex and more likely to overfit. Default: 5.
#         reg_lambda : float, optional
#             L2 regularization term on weights. Increasing this value will make model more conservative. Default: 1.0.
#         reg_alpha : float, optional
#             L1 regularization term on weights. Increasing this value will make model more conservative. Default: 0.1.

#         Returns
#         -------
#         torch.Tensor
#             A tensor of shape (batch_size, num_timesteps_output) representing the predicted values for the future timesteps.
#             Each element corresponds to a predicted value for a future timestep.
        
#     """
#     def __init__(self, num_features, num_timesteps_input, num_timesteps_output,
#                  n_estimators=40, learning_rate=0.1, max_depth=5, reg_lambda=1.0, reg_alpha=0.1):
#         self.num_features = num_features
#         self.num_timesteps_input = num_timesteps_input
#         self.num_timesteps_output = num_timesteps_output
#         self.n_estimators = n_estimators
#         self.learning_rate = learning_rate
#         self.max_depth = max_depth
#         self.reg_lambda = reg_lambda
#         self.reg_alpha = reg_alpha
#         self.model = None

#     def get_features(self, X):
#         features = X.reshape(-1, self.num_features * self.num_timesteps_input)
#         return features

#     def fit(self, train_input, train_target, val_input=None, val_target=None, epochs=1000, batch_size=10, verbose=False, patience=100):
#         """
#         Parameters
#         ----------
#         train_input : torch.Tensor or numpy.ndarray
#             The input training data, expected to be in the shape 
#             (batch_size, num_timesteps_input, num_features).
#         train_target : torch.Tensor or numpy.ndarray
#             The target training data, expected to be in the shape
#             (batch_size, num_timesteps_output).
#         val_input : torch.Tensor or numpy.ndarray, optional
#             The input validation data, following the same format as `train_input`.
#         val_target : torch.Tensor or numpy.ndarray, optional
#             The target validation data, following the same format as `train_target`.
#         epochs : int, optional
#             The number of epochs to train the model for. Default is 1000.
#         batch_size : int, optional
#             The size of batches to use when training the model. Default is 10.
#         verbose : bool, optional
#             If True, the training process will print out progress updates and evaluation metrics. Default is False.
#         patience : int, optional
#             Number of epochs with no improvement after which training will be stopped. Default is 100.

#         Notes
#         -----
#         The method first reshapes and prepares the input data for the XGBoost regressor, converting any PyTorch tensors
#         to numpy arrays if necessary. It configures the model with the specified hyperparameters and trains it on the
#         provided data. If validation data is provided, it will also evaluate the model on this data after each epoch,
#         which can be used for early stopping or parameter tuning. The method concludes by optionally printing the mean
#         squared error of the training and validation datasets if `verbose` is True.
#         """
#         train_input = self.get_features(train_input)
#         train_target = train_target.numpy() if isinstance(train_target, torch.Tensor) else train_target

#         if val_input is not None and val_target is not None:
#             val_input = self.get_features(val_input)
#             val_target = val_target.numpy() if isinstance(val_target, torch.Tensor) else val_target

#         self.model = xgb.XGBRegressor(
#             objective='reg:squarederror', 
#             n_estimators=self.n_estimators, 
#             learning_rate=self.learning_rate, 
#             max_depth=self.max_depth,
#             reg_lambda=self.reg_lambda,  
#             reg_alpha=self.reg_alpha    
#         )

#         if val_input is not None and val_target is not None:
#             self.model.fit(train_input, train_target, eval_set=[(train_input, train_target), (val_input, val_target)], verbose=verbose)
#         else:
#             self.model.fit(train_input, train_target)

#         # Optionally, print MSE if verbose
#         if verbose:
#             train_pred = self.model.predict(train_input)
#             train_mse = mean_squared_error(train_target, train_pred)
#             print("Training MSE: ", train_mse)
#             if val_input is not None and val_target is not None:
#                 val_pred = self.model.predict(val_input)
#                 val_mse = mean_squared_error(val_target, val_pred)
#                 print("Validation MSE: ", val_mse)

#     def predict(self, feature):
#         feature = self.get_features(feature)
#         prediction = self.model.predict(feature)

#         return torch.tensor(prediction)
